//
// DBaseCsv.C(PP)
//
// implementation of the CSV database file handling 
// functions.
//
// Martin Jansen
//

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>

#if OPH1004 || OPN2002
#include <sys\types.h>            // delete this line for Turbo C
#include <sys\stat.h>
#endif

#include <string.h>

#include "DBaseCsv.h"
#include "BPLUS.H"
#include "Database.h"

#ifdef WIN32
	#include <io.h>
#else
	//#include "WaitCurs.h"
#endif
#include "lib.h"

#if OPH1004 || OPN2002
#include <unistd.h>
#undef O_BINARY
#define O_BINARY 0x00
#endif

#ifdef WIN32
	#define open		_open
	#define close		_close
	#define read		_read
	#define write		_write
	#define lseek		_lseek
	#define O_RDWR		_O_RDWR 
	#define O_BINARY	_O_BINARY
	#define O_CREAT		_O_CREAT 
	#define O_TRUNC		_O_TRUNC  
#endif


extern char g_recordBuffer[];
extern char g_fieldBuffer[];

int CsvGetField( char *fieldBuffer, const char *buffer, int fieldNr, int maxFieldSize, int delimiter )
{
	int inQuote = 0;
	int fieldCount = 0;
	char ch;
	int fieldDataSize = 0;
	
	fieldBuffer[0] = '\0';
	for(;;)
	{
		if( buffer == NULL || *buffer == '\0' || *buffer == '\r' || *buffer == '\n' )
			return -100;	// HACK HACK HACK

		if( (inQuote = ( *buffer == '\"' )) != 0 )
			buffer++;	// Skip the quote

		if( fieldCount == fieldNr )
		{
			// Found the field nummer now copy the data
			// until the next delimiter or terminator
			while( ((ch = *buffer) != 0) && ch != '\r' && ch != '\n' && ((!inQuote && (*buffer != (char)delimiter )) || inQuote) )
			{
				if( ch == '\"' )
				{
					buffer++;
					inQuote = 0;
					continue;
				}
				*fieldBuffer++ = ch;
				fieldDataSize++;
				buffer++;
				if( fieldDataSize >= maxFieldSize )
				{
					*fieldBuffer = '\0';
					return fieldDataSize;
				}
			}
			*fieldBuffer = '\0';

			return fieldDataSize;	// HACK HACK HACK the length of the field
		}

		while( *buffer && ((!inQuote && (*buffer != (char)delimiter )) || inQuote))
		{
			if( *buffer == '\"' )
				inQuote = 0;
			buffer++;
		}
		if( *buffer == '\0' || *buffer == '\r' || *buffer == '\n' )
			return -1;
		buffer++;
		fieldCount++;
	}
}


int CsvBuildIndex( const char *dbName, const char *idxName, int delimiter, int fieldNr, int dup )
{
	static IX_DESC ix;
	static ENTRY e;
	FILE *fp;
	int error;
	long filePos = 0L;
	int fieldLength;
	//static char fieldBuffer[ MAX_G_FIELD + 1];
	// Open the database in binary read only mode

	if( (fp = fopen( dbName, "rb")) == NULL )
		return ERR_OPEN_DATABASE;

	if( MakeIndex( idxName, &ix, dup ) == IX_FAIL )
	{
		fclose( fp );
		return ERR_OPEN_INDEX; 
	}

	error = OK;

#ifdef OPH1004
	StartWaitCursor();
#endif

	while( fgets( g_recordBuffer, MAX_G_BUFFER, fp ) != NULL )
	{
		// parse the csv string and use the field number as new index value
		if( (fieldLength = CsvGetField( g_fieldBuffer, g_recordBuffer, fieldNr, MAX_G_FIELD, delimiter )) < 0 )
		{
			// empty record line or empty field
			// continue with indexing next record
			filePos = ftell( fp );
			continue;
		}
		if( fieldLength == 0 )
		{
			// empty record line or empty field
			// continue with indexing next record
			filePos = ftell( fp );
			continue;
		}
		
		e.key = hash(g_fieldBuffer);
		e.recptr  = (RECPOS)filePos;

		// Add the key to the indexing file
		if( AddKey( &e, &ix ) != IX_OK )
		{
			error = ERR_INDEX_DUPLICATE;
			goto build_index_error;
		}
		filePos = ftell( fp );
	}

build_index_error:
#ifdef OPH1004
	StopWaitCursor();
#endif
	fclose( fp );
	CloseIndex( &ix );
	return error;
}

int CsvRebuildIndex(  SDatabaseItem *dbItem, SDBOutVal *dbOut )
{
	int ret;
	CloseDatabase( dbOut );

	if( (ret = CsvBuildIndex( dbItem->dbFilename, dbItem->idxFilename, dbItem->csvChar, dbItem->keyfield, dbItem->flags & F_DB_APPEND )) < OK )
	{
		// Error while creating index file, so remove it
		remove( dbItem->idxFilename );
		return ret;	// ERROR!!!!
	}

	if( OpenIndex( dbItem->idxFilename, &(dbOut->ix), dbItem->flags & F_DB_APPEND ) != IX_OK )
		return ERR_OPEN_INDEX;

	dbOut->fdDb = open( dbItem->dbFilename, O_RDWR | O_BINARY, 0777 );
	return DATABASE_OK;
}

int CsvOpenDatabase( SDatabaseItem *dbItem, SDBOutVal *dbOut )
{
	int ret;

	dbOut->fdDb = open( dbItem->dbFilename, O_RDWR | O_BINARY | O_APPEND, 0777 );

	if( dbOut->fdDb < 0 )
		return ERR_OPEN_DATABASE;

	if( OpenIndex( dbItem->idxFilename, &(dbOut->ix), dbItem->flags & F_DB_APPEND ) != IX_OK )
	{
		// Close the original database first
		close( dbOut->fdDb );

		if( (ret = CsvBuildIndex( dbItem->dbFilename, dbItem->idxFilename, dbItem->csvChar, dbItem->keyfield, dbItem->flags & F_DB_APPEND )) < OK )
		{
			// Error while creating index file, so remove it
			remove( dbItem->idxFilename );
			return ret;	// ERROR!!!!
		}

		if( OpenIndex( dbItem->idxFilename, &(dbOut->ix), dbItem->flags & F_DB_APPEND ) != IX_OK )
			return ERR_OPEN_INDEX;

		dbOut->fdDb = open( dbItem->dbFilename, O_RDWR | O_BINARY | O_APPEND, 0777 );
	}
	dbOut->dbFlags = dbItem->flags;
	return DATABASE_OK;
}


int CsvCreateDatabase( SDatabaseItem *dbItem, SDBOutVal *dbOut )
{
	dbOut->fdDb = open( dbItem->dbFilename, O_RDWR | O_CREAT | O_TRUNC | O_BINARY, 0777);

   // dbOut->fdDb = open( dbItem->dbFilename, O_RDWR | O_CREAT | O_TRUNC | O_BINARY, 0777);
	if( dbOut->fdDb < 0 )
		return ERR_OPEN_DATABASE;

	if( MakeIndex( dbItem->idxFilename, &(dbOut->ix), dbItem->flags & F_DB_APPEND ) == IX_OK )
	{
		dbOut->dbFlags = dbItem->flags;
		return DATABASE_OK;
	}
	close( dbOut->fdDb );

	return ERR_OPEN_INDEX;	// Error creating the index file 
}


