//------------------------------------------------------------------
//  Batch mode application for the OPN2004
//------------------------------------------------------------------
//  This demo allows the OPN2004 to read barcodes, store them and
//  transmit them by OseComm using USB.
//
//  This application also support the reading of the
//  universal menu-book, so all sorts of barcode options can easily be tested.
//
//  Since all changed configurations are stored in flash, any changed
//  configurations are restored after a restart or when the battery was empty.
//
//  See the application manual on more details on the support
//  features and on how the application works.
//------------------------------------------------------------------
//  Version history:
//
//
//  RFI37931    :       Bugfix for MSD and removing barcodes.
//                      Minimal required OS version: RBIV0014b
//                      Release date: April 10th, 2013
//                      Author: Inanc Yigit
//                      Company: Opticon Sensors Europe BV
//
//  RFI37930	:		Ported to the OPN2004.
//                      Minimal required OS version: RBIV0010d
//						Release date: March 15th, 2013
//						Author: Inanc Yigit
//						Company: Opticon Sensors Europe BV
//
//------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "lib.h"
#include "main.h"
#include "database.h"
#include "debug.h"

#include <unistd.h>
#include <sys/types.h>

#include "menu_options.inc"		// Contains table with all application specific menu options
// Defines and variables used for barcode reading
#define	MAX_BARCODE_LENGTH			1000

// Error values
#define ILLEGAL_QUANTITY			0xFFFF
#define DATABASE_ERROR				0xFFFE
#define FILESYSTEM_ERROR			0xFFFD
#define ERROR_VALUE					0xFFF0

// Database record size
#define SZ_BARCODE					(MAX_BARCODE_LENGTH+1)
#define SZ_QUANTITY					9
#define SZ_SERIAL					6
#define SZ_TIME						(2+1+2+1+2)
#define SZ_DATE						(4+1+2+1+2)
#define SZ_DELETED					1
#define SZ_RECORD					(SZ_BARCODE+1+SZ_QUANTITY+1+SZ_SERIAL+1+SZ_TIME+1+SZ_DATE+1+SZ_DELETED+1+1)

#define FNAME_SCANNED				"SCANNED.TMP"
#define FNAME_SCANNED_IDX			"SCANNED.IDX"
#define FNAME_SCANNED_EXPORT		"SCANNED.TXT"

#define SZ_SCANNED					SZ_RECORD

// barcode record structure
typedef struct
{
	char deleted[SZ_DELETED + 1];
	char barcode[SZ_BARCODE + 1];
	char quantity[SZ_QUANTITY + 1];
	char serial[SZ_SERIAL + 1];
	char time[SZ_TIME + 1];
	char date[SZ_DATE + 1];
} db_record;

static char record[SZ_RECORD + 1];
static char bcr_buf[MAX_BARCODE_LENGTH];
static struct barcode code = {0};

// Locals
SDatabaseItem dbItemStorage =
{
		FNAME_SCANNED,							// char *dbFilename;
        FNAME_SCANNED_IDX,						// char *idxFilename
        F_DB_CSV | F_DB_APPEND,					// unsigned char flags
        SZ_SCANNED,								// int  maxRecordSize
        0,										// int  keyfield
        ','										// int csvChar
};
static SDBOutVal dbStorage = { -1 };

// Timer related stuff
#define MAX_TIMERS					1
#define FORMAT_TIMER				0			// Timer tracking clear key being held to erase stored data

#define IDLE_POWER_DOWN_TIME		2*50		// Power down time when idle: 2 seconds
#define IDLE_POWER_OFF_TIME			120*50		// Power off time when idle: 2 minutes
#define CLR_KEY_FORMAT_TIME			10*50		// Time clear key must be held to erase stored data

volatile unsigned int timer[MAX_TIMERS];		// Array of various timers

// Misc
int Aiming_Enabled = FALSE;
static int reset_read_mode;

//
// States of the main state machine of this application
//
#define STATE_IDLE					0			// Idle state
#define STATE_USB_CONNECTED			2			// Connected to USB
#define STATE_WAIT_FOR_RELEASE		3			// Wait until both keys are not pressed and then jump back idle state
//
// Public & local functions
//
void Application_Default(int comport);
void Version_Output(void);
static int GetBarcodeFieldNr(void);
static int GetQuantityFieldNr(void);
static int ShutdownCallback(void);
static int ExportDatabase(void);
//static int ImportDatabase(void);
int FieldSize(int field_type);

//
// Note: Only change this version if the application settings of the previous version are no longer compatible
//
#define	VERSION_CHECK				"BATCHDEMO_V5"	// Verification string used by 'LoadApplicationSettings' to check if any possible
													// application settings found in flash memory are compatible with this application.
													// If this version check fails all application settings will be set back to default
			
application_vars app; // Structure containing all application settings that will be stored in flash memory

static int last_key_pressed = -1; // Used to determine whether a barcode was read using the clear key or trigger key

//------------------------------------------------------------------
//  Application_Default
//  ============================
//  Reset all application specific settings to the specified default
//------------------------------------------------------------------
void Application_Default(int comport)
{
	switch(comport)
	{
		case COM11:
			app.default_interface = COM11;
			systemsetting("C04");  // Switch OS to MSD mode
			break;

		case COM0:
		case COM9:
		default:
			app.default_interface = COM0;
			systemsetting("C01");   // Switch to Terminal default (USB-VCP/RS232)
			break;
	}

	systemsetting("YC");											// Enable menu-labels
	systemsetting("7G");											// Enable charge indicator
	systemsetting("S0");											// Single read
	systemsetting("Y0");											// Trigger mode

	app.goodread_duration = 10;										// Good read led duration (200ms)
	app.buzzer_volume = VHIGH;										// Buzzer volume (high)
	app.buzzer_duration = TSTANDARD;								// Buzzer duration (200ms)
    app.buzzer_options = BUZZER_ENABLED;							// Buzzer enabled, no startup buzzer enabled
	app.buzzer_options |= BUZZER_TONE_HIGH_LOW;						// Buzzer tone (High-low tone)
	app.special_options = OSECOMM_COMMUNICATE;						// Use OseComm by default
	app.read_mode = SINGLE;                                         // Single read
    app.read_time = 0;                                              // 0 seconds (trigger)
	app.func_key_del = FALSE;										// Disable ability to format flash disk by holding function key for CLR_KEY_FORMAT_TIME

	app.output_field_separator = ',';								// Output field separator
	app.output_date_time_format = TIME_FORMAT_HH_MM_SS;				// Output time format hh:mm:ss
	app.output_date_time_format |= DATE_FORMAT_SLASH_DD_MM_YYYY;	// Output data dd/mm/yyyy
	app.output_barcode_size = 0x00;									// Output barcode field size (variable length fields)
	app.output_quantity_size = 0x00;								// Output quantity field size (0=empty field)
	app.output_quantity_options = QUANTITY_DELETE_ZERO_QNTY;		// Don't allow negative quantity, delete key enabled, delete record is quantity reaches zero

	app.output_sequence[0] = SEQUENCE_BARCODE;						// Output sequence: Barcode, Quantity, Serial number, Time, Date, Deleted
	app.output_sequence[1] = SEQUENCE_QUANTITY;
	app.output_sequence[2] = SEQUENCE_SERIAL_NUMBER;
	app.output_sequence[3] = SEQUENCE_TIME;
	app.output_sequence[4] = SEQUENCE_DATE;
}

//------------------------------------------------------------------
//  Version_Output
//  ============================
//  Output all versions, including the application version
//------------------------------------------------------------------
void Version_Output(void)
{
	systemsetting("Z1");
	printf("%s\r", (char *)VERSION);
}

//------------------------------------------------------------------
//  SetReadMode
//  ============================
//  Converts the configured read time/mode to a scannerpower call
//------------------------------------------------------------------
void SetReadMode(int readmode, int readtime)
{
    switch (readmode)
    {
    case MULTIPLE:
        if (readtime == 0)
            scannerpower(MULTIPLE | TRIGGER, 10*50);        // Limit trigger mode to 10 seconds to avoid pressing the trigger for 20s, causing an unintended restart
        else if (readtime == READ_TIME_INDEFINITELY)
            scannerpower(MULTIPLE, -1);
        else
            scannerpower(MULTIPLE, readtime*5);
        break;

    case CONTINUOUS:
        if (readtime == 0)
            scannerpower(CONTINUOUS | TRIGGER, 10*50);      // Limit trigger mode to 10 seconds to avoid pressing the trigger for 20s, causing an unintended restart
        else if (readtime == READ_TIME_INDEFINITELY)
            scannerpower(CONTINUOUS, -1);
        else
            scannerpower(CONTINUOUS, readtime*5);
        break;

    case SINGLE:
    default:
        if (readtime == 0)
            scannerpower(SINGLE | TRIGGER, 10*50);          // Limit trigger mode to 10 seconds to avoid pressing the trigger for 20s, causing an unintended restart
        else if (readtime == READ_TIME_INDEFINITELY)
            scannerpower(SINGLE, -1);
        else
            scannerpower(SINGLE, readtime*5);
        break;
    }
}

//------------------------------------------------------------------
//  OnCommInfo
//  ============================
//  Callback function that gets the status, error and progress
//  information from 'OseComm' during the up- and dowloading of files.
//  Visual process information is shown by this callback function.
//------------------------------------------------------------------
void OnCommInfo(int status, int errorsuccess, int progress, const char *info)
{
	switch (status)
	{
		case STAT_LIST_FILES:
			break;
			
		case STAT_XMIT_FILE_TO_PC:
			if (errorsuccess == SUCC_COMPLETE)
			{
				sound(TSTANDARD, VHIGH, SLOW, SMEDIUM, SHIGH, SLOW, SMEDIUM, SHIGH, 0);		// Ready beep
				remove(FNAME_SCANNED);
				remove(FNAME_SCANNED_EXPORT);
				remove(FNAME_SCANNED_IDX);		// Remove current link index file as well
			}
			break;
			
		case STAT_RECV_FILE_FROM_PC:
		case STAT_GET_TIME_DATE:
		case STAT_SET_TIME_DATE:
		case STAT_GET_OS_VERSION:
		case STAT_GET_APPL_VERSION:
			break;
	}
}

//------------------------------------------------------------------
//  Communicate
//  ============================
//
//------------------------------------------------------------------
void Communicate(void)
{
    char serial_number[6+1];
	int result;
	
	if (app.special_options == OSECOMM_COMMUNICATE)
		OseComm(-1L, CRADLE_ABORT, VERSION_CHECK, OnCommInfo);
	else
	{
		if( fsize((char*)FNAME_SCANNED_EXPORT) == -1L )
		{
			return;
		}

		sprintf(serial_number, "%-6.6lu", GetSerialNumber());

		result = neto_transmit( (char(*)[MAX_FNAME])FNAME_SCANNED_EXPORT, 1, serial_number, TRIGGER_KEY, 3);

		if(result >= 0)
		{
			remove(FNAME_SCANNED);
			remove(FNAME_SCANNED_EXPORT);
			remove(FNAME_SCANNED_IDX);		// Remove current link index file as well
			sound( TSTANDARD, VHIGH, SMEDIUM, SHIGH, SLOW, SMEDIUM, SHIGH, 0 );		// Success beep
		}
	}
}

//------------------------------------------------------------------
//  Timers_Heartbeat
//  ============================
//  Heartbeat routine that is executed by the OS every 20ms, causing
//  the seperate timers to be decremented (if enabled)
//------------------------------------------------------------------
void Timers_Heartbeat( void )
{
	for( int x = 0; x < MAX_TIMERS; x++ )
	{
		if( timer[x] != 0 && timer[x] != 0xFFFF )
			timer[x]--;
	}
}

//------------------------------------------------------------------
//  StoreBarcodeData
//  ============================
//  Saves the barcode data into the database file
//------------------------------------------------------------------
int StoreBarcodeData(db_record *db_rec, long lRecordNo)
{
	char *ptr;
	int i;
	int barcode_field, quantity_field;
	int field_size;
	
	// Check if there is enough space available for storing the barcode data
	if (coreleft() < 1024)
	{
		return FILESYSTEM_ERROR;
	}

	barcode_field = GetBarcodeFieldNr();

	if (barcode_field == -1)
		return ERROR;

	quantity_field = GetQuantityFieldNr();
	
	// Start building the output record
	ptr = record;

	// Add deleted field
	sprintf(ptr, "%s%c", db_rec->deleted, app.output_field_separator);
	ptr += strlen(db_rec->deleted) + 1;

	// Add remaining fields as configured
	for (i = 0; i < SEQUENCE_MAX; ++i)
	{
		field_size = FieldSize(app.output_sequence[i]);
		
		if ((app.output_sequence[i] == SEQUENCE_BARCODE) && (app.output_barcode_size == 0x00))
		{
			field_size = strlen(db_rec->barcode);
		}
		else if ((app.output_sequence[i] == SEQUENCE_QUANTITY) && (app.output_quantity_size == 0x00))
		{
			field_size = strlen(db_rec->quantity);
		}
		
		switch (app.output_sequence[i])
		{
			case SEQUENCE_BARCODE:
				//strcpy(ptr, db_rec->barcode);
				sprintf(ptr, "%-*.*s", field_size, field_size, db_rec->barcode);
				//ptr += strlen(db_rec->barcode);				// Skip the field
				break;
				
			case SEQUENCE_QUANTITY:
				//strcpy(ptr, db_rec->quantity);
				sprintf(ptr, "%-*.*s", field_size, field_size, db_rec->quantity);
				//ptr += strlen(db_rec->quantity);
				break;
				
			case SEQUENCE_SERIAL_NUMBER:
				//strcpy(ptr, db_rec->serial);
				sprintf(ptr, "%-*.*s", field_size, field_size, db_rec->serial);
				//ptr += strlen(db_rec->serial);
				break;
				
			case SEQUENCE_TIME:
				//strcpy(ptr, db_rec->time);
				sprintf(ptr, "%-*.*s", field_size, field_size, db_rec->time);
				//ptr += strlen(db_rec->time);
				break;
				
			case SEQUENCE_DATE:
				//strcpy(ptr, db_rec->date);
				sprintf(ptr, "%-*.*s", field_size, field_size, db_rec->date);
				//ptr += strlen(db_rec->date);
				break;
		}
		
		ptr += field_size;
		
		if (i == (SEQUENCE_MAX - 1) || app.output_sequence[i + 1] == SEQUENCE_NONE || app.output_sequence[i + 1] > SEQUENCE_MAX)
		{
			sprintf(ptr, "\r\n");
			break;
		}

		*ptr++ = app.output_field_separator;	// Add field separator
	}
	
	dbItemStorage.keyfield = barcode_field;
	
	dbItemStorage.flags = (quantity_field == -1) ? (F_DB_CSV | F_DB_APPEND) : (F_DB_CSV);

	dbItemStorage.csvChar = app.output_field_separator;

	if (WriteRecord(&dbItemStorage, &dbStorage, &lRecordNo, record, 0) != DATABASE_OK)
	{
		return FILESYSTEM_ERROR;
	}
	return OK;
}

//------------------------------------------------------------------
//  StringQuantityToLong
//  ============================
//  Convert the stored quantity string to a long value
//------------------------------------------------------------------
long StringQuantityToLong(char* quantity)
{
	static char tmp[SZ_QUANTITY + 1];
	char* ptr;
	int pos;

	memset(tmp, '\0', sizeof(tmp));
	ptr = tmp;
	for (pos = 0; pos < (SZ_QUANTITY); pos++)
	{
		if (quantity[pos] == '\0')
			break;
		
		if (quantity[pos] == ' ') // cut off all spaces
			continue;
		
		*ptr++ = quantity[pos];
	}
	if (atol(tmp) == 0L && tmp[0] == '-')
		return 0;
	
	return (atol(tmp));
}

//------------------------------------------------------------------
//  FieldSize
//  ============================
//  Returns the string length of the specified field type
//------------------------------------------------------------------
int FieldSize(int field_type)
{
	switch (field_type)
	{
		case SEQUENCE_BARCODE:	
			return (app.output_barcode_size>>4)*10 + (app.output_barcode_size & 0xF);
			
		case SEQUENCE_QUANTITY:
			return app.output_quantity_size;
			
		case SEQUENCE_SERIAL_NUMBER:
			return SZ_SERIAL;

		case SEQUENCE_TIME:
			if ((app.output_date_time_format & TIME_FORMAT_MASK) == TIME_FORMAT_HH_MM)
				return 5;

			return 8;		// TIME_FORMAT_HH_MM_SS:

		case SEQUENCE_DATE:

			switch (app.output_date_time_format & DATE_FORMAT_MASK)
			{
				case DATE_FORMAT_DASH_DD_MM_YYYY:
				case DATE_FORMAT_DASH_MM_DD_YYYY:
				case DATE_FORMAT_DASH_YYYY_MM_DD:
				case DATE_FORMAT_SLASH_DD_MM_YYYY:
				case DATE_FORMAT_SLASH_MM_DD_YYYY:
				case DATE_FORMAT_SLASH_YYYY_MM_DD:
					return 10;

				case DATE_FORMAT_DASH_DD_MM_YY:
				case DATE_FORMAT_DASH_MM_DD_YY:
				case DATE_FORMAT_DASH_YY_MM_DD:
				case DATE_FORMAT_SLASH_DD_MM_YY:
				case DATE_FORMAT_SLASH_MM_DD_YY:
				case DATE_FORMAT_SLASH_YY_MM_DD:
					return 8;

				case DATE_FORMAT_DASH_DD_MM:
				case DATE_FORMAT_DASH_MM_DD:
				case DATE_FORMAT_SLASH_DD_MM:
				case DATE_FORMAT_SLASH_MM_DD:
					return 5;
			}
			break;
	}
	
	return 0;
}

//-------------------------------------------------------------------
//  FindBarcodeInDatabase
//  ============================
//  Searches for the specified barcode in the database and returns it
//-------------------------------------------------------------------
long FindBarcodeInDatabase(char *barcode, char *quantity)
{
	static char found_barcode[SZ_BARCODE+1];

	long lFound = -1L;
	long lFoundSec = -1;
	int barcode_field, quantity_field;
	int barcode_length = strlen(barcode);

	barcode_field = GetBarcodeFieldNr();

	if (barcode_field == -1)
		return -1L;
	
	quantity_field = GetQuantityFieldNr();
	
	dbItemStorage.keyfield = barcode_field;

	dbItemStorage.csvChar =	app.output_field_separator;

	dbItemStorage.flags = (quantity_field == -1) ? (F_DB_CSV | F_DB_APPEND) : (F_DB_CSV);	// Allow same barcode twice or not

	if (quantity_field == -1)
	{
		if (last_key_pressed != CLR_KEY)		// Little bit ugly speed hack to avoid doing a slow database search when it's not even needed
		{
			return -1L;							// Return -1, so record will be appended
		}

		if (SearchDatabase(&dbItemStorage, &dbStorage, barcode, &lFound, record) != DATABASE_OK)
		{
			return -1L;							// Barcode not found
		}

		//
		// Found a code but maybe there is an newer code available
		//
		for (;;)
		{
			if (ReadPreviousRecord(&dbItemStorage, &dbStorage, &lFoundSec, record) != DATABASE_OK)
				break;

			// Compare the barcode
			if( GetField(&dbItemStorage, found_barcode, record, barcode_field, SZ_BARCODE) != barcode_length)
				break;

			if (memcmp(found_barcode, barcode, barcode_length) != 0)
	            break;

			if (record[0] == 'X')	// Code already marked deleted
				continue;

			// Newer barcode found
			lFound = lFoundSec;
		}

		//
	    // Read newest matching barcode again

		if (ReadRecord(&dbItemStorage, &dbStorage, lFound, record) != DATABASE_OK)
		{
			return -1L;
		}

		if (record[0] == 'X') // Found code is already deleted
		{
			// Search for older barcode not marked deleted
			for (;;)
			{
				if (ReadNextRecord(&dbItemStorage, &dbStorage, &lFoundSec, record) != DATABASE_OK)
				{
					// No older code in storage
					return -1L;
				}

				// Compare the barcode
				if( GetField(&dbItemStorage, found_barcode, record, barcode_field, SZ_BARCODE) != barcode_length)
				{
					// No older code in storage
					return -1L;
				}

				if (memcmp(found_barcode, barcode, barcode_length) != 0)
				{
					// No older code in storage
					return -1L;
				}

				if (record[0] == 'O')
				{
					// Found an older barcode
					lFound = lFoundSec;
					break;
				}
			}
        }
	}
	else
	{
		if (SearchDatabase(&dbItemStorage, &dbStorage, barcode, &lFound, record) == DATABASE_OK)
		{
			// Fill the quantity string
			GetField(&dbItemStorage, quantity, record, quantity_field, SZ_QUANTITY);
		}
	}
	return lFound;
}

//------------------------------------------------------------------
//  StoreBarcode
//  ============================
//  Stores a barcode in the database
//------------------------------------------------------------------
int StoreBarcode(struct barcode *code, int amount)
{
	static char quantity[SZ_QUANTITY + 1];
	static db_record db_rec;
	struct date dates;
	struct time times;
	long lFoundRecord;
	long lTotal;
	int do_delete = FALSE, result;
	int	barcode_size = FieldSize(SEQUENCE_BARCODE);

	memset(&db_rec, '\0', sizeof(db_record));
	
	if ((barcode_size == 0x00) && (app.output_barcode_size == 0x00))
		barcode_size = strlen(code->text);

	// fill the barcode into the record structure
	//strcpy(db_rec.barcode, code->text);
	sprintf( db_rec.barcode, "%-*.*s", barcode_size, barcode_size, code->text );

	memset(quantity, '\0', sizeof(quantity));	// clear the whole quantity item

	if ((lFoundRecord = FindBarcodeInDatabase(db_rec.barcode, quantity)) != -1L)
	{
		if (strlen(quantity) == 0 && amount < 0)	// If a record was found, but no quantity field
		{
			do_delete = TRUE;
			lTotal = 0;
		}
		else
		{

			lTotal = StringQuantityToLong(quantity);

			if (!(app.output_quantity_options & QUANTITY_ALLOW_NEGATIVE_QNTY))
			{
				if (lTotal == 0 && amount < 0)
					return ILLEGAL_QUANTITY;
			}

			lTotal += amount;

			if (lTotal == 0 && (app.output_quantity_options & QUANTITY_DELETE_ZERO_QNTY))
			{
				do_delete = TRUE;
			}

		}
	}
	else
	{
		if (!(app.output_quantity_options & QUANTITY_ALLOW_NEGATIVE_QNTY))
		{
			if (amount < 0)
			{
				return ILLEGAL_QUANTITY;
			}

		}

		lTotal = amount;
	}

	// When ok fill the quantity of the db_record
	//sprintf(db_rec.quantity, "%ld", lTotal);
	sprintf(db_rec.quantity, "%*ld", app.output_quantity_size, lTotal);

	sprintf(db_rec.serial, "%-6.6lu", GetSerialNumber());

	gettime(&times);

	switch (app.output_date_time_format & TIME_FORMAT_MASK)
	{
		case TIME_FORMAT_HH_MM:
			sprintf(db_rec.time, "%02d:%02d", times.ti_hour, times.ti_min);
			break;

		case TIME_FORMAT_HH_MM_SS:
			sprintf(db_rec.time, "%02d:%02d:%02d", times.ti_hour, times.ti_min, times.ti_sec);
			break;
	}

	getdate(&dates);

	switch (app.output_date_time_format & DATE_FORMAT_MASK)
	{
		case DATE_FORMAT_SLASH_DD_MM_YYYY:
			sprintf(db_rec.date, "%02d/%02d/%04d", dates.da_day, dates.da_mon, dates.da_year);
			break;
		case DATE_FORMAT_DASH_DD_MM_YYYY:
			sprintf(db_rec.date, "%02d-%02d-%04d", dates.da_day, dates.da_mon, dates.da_year);
			break;
		case DATE_FORMAT_SLASH_MM_DD_YYYY:
			sprintf(db_rec.date, "%02d/%02d/%04d", dates.da_mon, dates.da_day, dates.da_year);
			break;
		case DATE_FORMAT_DASH_MM_DD_YYYY:
			sprintf(db_rec.date, "%02d-%02d-%04d", dates.da_mon, dates.da_day, dates.da_year);
			break;
		case DATE_FORMAT_SLASH_YYYY_MM_DD:
			sprintf(db_rec.date, "%04d/%02d/%02d", dates.da_year, dates.da_mon, dates.da_day);
			break;
		case DATE_FORMAT_DASH_YYYY_MM_DD:
			sprintf(db_rec.date, "%04d-%02d-%02d", dates.da_year, dates.da_mon, dates.da_day);
			break;
		case DATE_FORMAT_DASH_DD_MM_YY:
			sprintf(db_rec.date, "%02d-%02d-%02d", dates.da_day, dates.da_mon, dates.da_year % 100);
			break;
		case DATE_FORMAT_SLASH_DD_MM_YY:
			sprintf(db_rec.date, "%02d/%02d/%02d", dates.da_day, dates.da_mon, dates.da_year % 100);
			break;
		case DATE_FORMAT_SLASH_MM_DD_YY:
			sprintf(db_rec.date, "%02d/%02d/%02d", dates.da_mon, dates.da_day, dates.da_year % 100);
			break;
		case DATE_FORMAT_DASH_MM_DD_YY:
			sprintf(db_rec.date, "%02d-%02d-%02d", dates.da_mon, dates.da_day, dates.da_year % 100);
			break;
		case DATE_FORMAT_SLASH_YY_MM_DD:
			sprintf(db_rec.date, "%02d/%02d/%02d", dates.da_year % 100, dates.da_mon, dates.da_day);
			break;
		case DATE_FORMAT_DASH_YY_MM_DD:
			sprintf(db_rec.date, "%02d-%02d-%02d", dates.da_year % 100, dates.da_mon, dates.da_day);
			break;
		case DATE_FORMAT_SLASH_MM_DD:
			sprintf(db_rec.date, "%02d/%02d", dates.da_mon, dates.da_day);
			break;
		case DATE_FORMAT_DASH_MM_DD:
			sprintf(db_rec.date, "%02d-%02d", dates.da_mon, dates.da_day);
			break;
		case DATE_FORMAT_SLASH_DD_MM:
			sprintf(db_rec.date, "%02d/%02d", dates.da_day, dates.da_mon);
			break;
		case DATE_FORMAT_DASH_DD_MM:
			sprintf(db_rec.date, "%02d-%02d", dates.da_day, dates.da_mon);
			break;
	}
	
	sprintf(db_rec.deleted, (do_delete) ? "X" : "O");

	// Store the input data into our database
	result = StoreBarcodeData(&db_rec, lFoundRecord);

	if (result != OK)
		return result;

	return (int) lTotal;
}

//------------------------------------------------------------------
//  OkBeep
//  ============================
//  Sounds a good read beep according to the current configuration
//------------------------------------------------------------------
void OkBeep(int base_frequency)
{
	unsigned char length, volume;

	if (!(app.buzzer_options & BUZZER_ENABLED))
		return;

	volume = app.buzzer_volume;
	length = app.buzzer_duration;

	switch (app.buzzer_options & BUZZER_TONE_MASK)
	{
		case BUZZER_TONE_SINGLE:
			sound(length * 2, volume, base_frequency, 0);
			break;

		case BUZZER_TONE_LOW_HIGH:
			sound(length, volume, base_frequency, base_frequency + 1000, 0);
			break;

		case BUZZER_TONE_HIGH_LOW:
		default:
			sound(length, volume, base_frequency, base_frequency - 500, 0);
			break;
	}
	delay(10);
}

//------------------------------------------------------------------
//  ErrBeep
//  ============================
//  Sounds an error beep according to the current configuration
//------------------------------------------------------------------
void ErrBeep(void)
{
	unsigned char length, volume;

	volume = app.buzzer_volume;
	length = app.buzzer_duration;

	sound(length * 2, volume, SERROR, 0);
}

//------------------------------------------------------------------
//  GoodReadLed
//  ============================
//  Shows a good read LED according to the current configuration
//------------------------------------------------------------------
void GoodReadLed(int color)
{
	goodreadled(color, app.goodread_duration);
}

//------------------------------------------------------------------
//  ProcessBarcode
//  =====================
//  Processes a barcode that has been read. This function
//  makes is possible to read set-up labels from the universal
//  book as well. (Changed settings will be lost after a restart)
//------------------------------------------------------------------
int ProcessBarcode(struct barcode *code)
{
	int result = 0;

	//
	// Little dirty tweak for the PX20, which doesn't support Code-128 menu labels. 
	// To make it possible to read the set-up sheet at the back of the application manual
	// the following code has been added to read 2, 3 and 4 digit menu options
	//
	if (code->id == CODE128 && (code->length == 2 || code->length == 3 || (code->length == 4 && code->text[0] == '[') || (code->length == 5 && code->text[0] == ']')))
	{
		code->id = MENU_CODE_C128;
	}
	//


	if (code->id == MENU_CODE)	// Check for standard Code-39 menu labels (from the OSE Universal menu book)
	{
		while (result != EXITING_MENU_MODE && result != ERROR)
		{
			if (result == 0 || readbarcode(code) == OK)
			{
				switch ((result = ExecuteMenuLabel(code)))
				{
					case ENTERING_MENU_MODE:
						sound(TSTANDARD, VHIGH, SHIGH, SMEDIUM, SHIGH, 0);
						break;

					case EXITING_MENU_MODE:
						sound(TSTANDARD, VHIGH, SHIGH, SMEDIUM, SHIGH, 0);
						delay(TSTANDARD * 5);	// Wait till buzzer sound if finished before saving (saving flash settings turns off sound timer temporarily)
						systemsetting("Z2");		// Save changed settings in flash memory and apply possibly changed interface

						break;

					case INVALID_OPTION_READ:
						sound(TLONG, VHIGH, SLOW, 0);
						break;

					case VALID_OPTION_READ:
						sound(TSTANDARD, VHIGH, SHIGH, SMEDIUM, SHIGH, 0);
						break;

					case LABEL_IGNORED:
						break;

					case ERROR:
						break;
				}
			}
		}

		return MENU_CODE;
	}
	else if (code->id == MENU_CODE_PDF || code->id == MENU_CODE_C128 || code->id == MENU_CODE_QR)	// Check for Code-128 / QR / PDF417 menu labels
	{
		switch ((result = ExecuteMenuLabel(code)))
		{
			case INVALID_OPTION_READ:
				sound(TLONG, VHIGH, SLOW, 0);
				break;

			case VALID_OPTION_READ:
				sound(TSTANDARD, VHIGH, SMEDIUM, SHIGH, SMEDIUM, SHIGH, 0);
				delay(TSTANDARD * 5);	// Wait till buzzer sound if finished before saving (saving flash settings turns off sound timer temporarily)
				systemsetting("Z2");	// Save changed settings in flash memory and apply possibly changed interface
				break;

			case LABEL_IGNORED:
			default:
				break;
		}

		if(Aiming_Enabled)
			AimingOn();
		else
			AimingOff();
		
		return MENU_CODE;
	}

	//
	// Store barcode
	//
	if (last_key_pressed == CLR_KEY && !(app.output_quantity_options & QUANTITY_DELETE_KEY_DISABLED)) // If the CLR_KEY is pressed
	{
		result = StoreBarcode(code, -1);	// Substract 1 from database (if present/allowed)

		if (result >= ERROR_VALUE)
		{
			GoodReadLed(RED);
			ErrBeep();
		}
		else
		{
			if (result >= 0)
			{
				GoodReadLed(ORANGE); // ORANGE led indicates a succesfully deleted/substracted barcode
				OkBeep(2000);		// Barcode deleted, quantity >= 0 : 2000 Hz
			}
			else
			{
				GoodReadLed(RED);		// RED led indicates a succesfully substracted barcode, resulting in a negative value
				OkBeep(1500);		// Barcode deleted, quantity < 0 : 1500 Hz
			}
		}
	}
	else  // If the trigger is pressed
	{
		result = StoreBarcode(code, 1);		// Add 1 to database

		if (result >= ERROR_VALUE)
		{
			GoodReadLed(ORANGE);
			ErrBeep();
		}
		else
		{
			if (result >= 0)
			{
				GoodReadLed((result > 0) ? GREEN : ORANGE);
				OkBeep(3000);		// Barcode added, quantity >= 0 : 3000 Hz
			}
			else
			{
				GoodReadLed(RED);		// RED led indicates a succesfully added barcode, but still resulting in a negative value
				OkBeep(1500);		// Barcode added, quantity < 0 : 1500 Hz
			}
		}
	}
	
	return FALSE;
}

//------------------------------------------------------------------
//  EnableAiming
//  =====================
//  Enables aiming help for reading the set-up sheet of the manual
//------------------------------------------------------------------
int EnableAiming(void)
{
	Aiming_Enabled = TRUE;
	return OK;
}

//------------------------------------------------------------------
//  DisableAiming
//  =====================
//  Disable aiming help for reading the set-up sheet of the manual
//------------------------------------------------------------------
int DisableAiming(void)
{
	Aiming_Enabled = FALSE;
	return OK;
}

//------------------------------------------------------------------
//  UsbSetVcp
//  =====================
//  Select USB VCP Interface (and RS232 when placed in cradle)
//------------------------------------------------------------------
int UsbSetVcp(void)
{
	Application_Default(COM0);
	return OK;
}

//------------------------------------------------------------------
//  UsbSetMsd
//  =====================
//  Select USB MSD Interface
//------------------------------------------------------------------
int UsbSetMsd(void)
{
	Application_Default(COM11);
	return OK;
}

//------------------------------------------------------------------------------
//	GetBarcodeFieldNr
//	=====================
//
//------------------------------------------------------------------------------
int GetBarcodeFieldNr(void)
{
	int i, barcode_field;

	barcode_field = -1;
	
	for (i = 0; i < SEQUENCE_MAX; ++i)
	{
		if (app.output_sequence[i] == SEQUENCE_BARCODE)
		{
			barcode_field = i+1;
			break;
		}
		
		if (i == SEQUENCE_MAX - 1 || app.output_sequence[i + 1] == SEQUENCE_NONE || app.output_sequence[i + 1] > SEQUENCE_MAX)
			break;
	}
	
	return barcode_field;
}

//------------------------------------------------------------------------------
//	GetQuantityFieldNr
//	=====================
//
//------------------------------------------------------------------------------
int GetQuantityFieldNr(void)
{
	int i, quantity_field;

	quantity_field = -1;
	
	for (i = 0; i < SEQUENCE_MAX; ++i)
	{
		if (app.output_sequence[i] == SEQUENCE_QUANTITY)
		{
			quantity_field = i+1;
			break;
		}
		
		if (i == SEQUENCE_MAX - 1 || app.output_sequence[i + 1] == SEQUENCE_NONE || app.output_sequence[i + 1] > SEQUENCE_MAX)
			break;
	}
	
	return quantity_field;
}


//------------------------------------------------------------------------------
//	OpenStorage
//	=====================
//
//------------------------------------------------------------------------------
int OpenStorage(void)
{
	int barcode_field, quantity_field;
	
	if (dbStorage.fdDb >= 0)
	{
		return ERROR;
	}

	barcode_field = GetBarcodeFieldNr();

	if (barcode_field == -1)
		return ERROR;

	quantity_field = GetQuantityFieldNr();

	dbItemStorage.keyfield = barcode_field;

	dbItemStorage.csvChar = app.output_field_separator;

	dbItemStorage.flags = (quantity_field == -1) ? (F_DB_CSV | F_DB_APPEND) : (F_DB_CSV);	// Allow same barcode twice or not

	dbItemStorage.maxRecordSize = sizeof(record);
	
	// ImportDatabase();	// This would allows editing of SCANNED.TXT in USB-MSD mode, but the creation of the new index file gets very slow with large number of barcodes due to sorting

	if ( OpenDatabase(&dbItemStorage, &dbStorage) != DATABASE_OK)
	{
		// If database doesnt exist, create one
		if (CreateDatabase(&dbItemStorage, &dbStorage) != DATABASE_OK)
		{
			return ERROR;
		}
	}

	return OK;
}
//------------------------------------------------------------------------------
//	CloseStorage
//	=====================
//
//-----------------------------------------------------------------------------
int CloseStorage(void)
{
	if (dbStorage.fdDb < 0)
	{
		return ERROR;
	}

	CloseDatabase(&dbStorage);
	dbStorage.fdDb = -1;

	return OK;
}

//------------------------------------------------------------------------------
//	ShutdownCallback
//  =====================
// 	Closes the database file before shutting down
//------------------------------------------------------------------------------
int ShutdownCallback(void)
{
	CloseStorage();

	if (UsbStatus() != USB_NOT_CONNECTED && app.default_interface == COM11)	// If MSD -> Generate the export file here on connecting the USB-cable
    {
		ExportDatabase();					// Export the database
		goodreadled(GREEN_FLASH, -1);		// Show blinking green LED when connected to USB
    }

	return OK;
}



//------------------------------------------------------------------------------
// 	ExportDatabase
//  =====================
// 	Exports the database from the SCANNED.TMP file to the SCANNED.TXT file.
//	The .TMP file is for internal use. The .TXT file is for export to the computer.
// 	It removes the first column from the db. With all the 'X' and 'O''s.
//	It's builded due to speed issues. A copy database is faster than remove
// 	within a database.
//------------------------------------------------------------------------------
int ExportDatabase(void)
{
    FILE *dbIn = NULL, *dbOut = NULL;
	int res = 0;

    // Open the input file
    if (fsize(FNAME_SCANNED) <= 0 || (dbIn = fopen(FNAME_SCANNED, "rb")) == NULL)
    {
		remove(FNAME_SCANNED_IDX);		// Remove current link index file
        remove(FNAME_SCANNED);			// Remove temporary database
        return ERR_OPEN_DATABASE;
    }

	// Open the export file
    if ((dbOut = fopen(FNAME_SCANNED_EXPORT, "w+b")) == NULL)
    {
        // close database
        fclose(dbIn);
        return ERR_OPEN_DATABASE;	// return error opening database
    }

	goodreadled(ORANGE, -1);

    // while the next record exists
    while (fgets(record, sizeof(record), dbIn) != NULL)
    {
        // if it is not starting with X, so if you don't want to remove.
        if (record[0] != 'X')
        {
            // Copy till the end of the file.
            if ((res=fputs(record + 2, dbOut)) == EOF)
				break;
        }
    }

    // close database
    fclose(dbOut);
    fclose(dbIn);

	goodreadled(ORANGE, 0);

	if(res == EOF)
	{
		return ERR_DB_WRITE;	// return error writing DB
	}

    return OK;
}

//------------------------------------------------------------------------------
// 	ImportDatabase
//  =====================
// 	Imports the database from the .TXT file to the .TMP file
//	The .TMP file is for internal use. The .TXT file is for export to the computer.
// 	It adds the first column from the db. With all the 'X' and 'O''s.
//	This is needed when you open the device as a mass storage device and edit
//	the database.
//------------------------------------------------------------------------------
int ImportDatabase(void)
{
    FILE *dbIn = NULL, *dbOut = NULL;

	if (fsize(FNAME_SCANNED_EXPORT) <= 0 || (dbIn = fopen(FNAME_SCANNED_EXPORT, "rb")) == NULL)
    {
		remove(FNAME_SCANNED_EXPORT);		// Just to be sure
        return ERR_OPEN_DATABASE;
    }

	if ((dbOut = fopen(FNAME_SCANNED, "w+b")) == NULL)
    {
        fclose(dbIn);
        return ERR_OPEN_DATABASE;
    }

    while (fgets(record+2, sizeof(record)-2, dbIn) != NULL)
    {
		record[0] = 'O';							// paste "O" and 'output field seperator' in front of record
		record[1] = app.output_field_separator;
        
		if (fputs(record, dbOut)  == EOF)	
        {
            fclose(dbIn);
            fclose(dbOut); 
            return ERR_DB_WRITE;
        }
    }

    fclose(dbIn);
    fclose(dbOut);

	remove(FNAME_SCANNED_IDX);			// Remove old index file, if present, because it might be outdated
	remove(FNAME_SCANNED_EXPORT);		// Remove the exported database when done, it will be generated again when USB is connected
    return OK;
}

//------------------------------------------------------------------
//  main
//  =====================
//  Main function of this demo that allows the OPN2004 to read
//  barcodes and transmit them using the mini USB cable to a virtual
//  comport on a PC or laptop. This application also support the
//  reading of the univeral menu-book. Since all changed
//  configurations are stored in flash, the changed configurations
//  are restored after a restart even if the battery has been empty.
//------------------------------------------------------------------
void main(void)
{
	int state, ch;

	// Initialize the barcode structure
	code.text = bcr_buf;
	code.min = 0;
	code.max = MAX_BARCODE_LENGTH;

	state = STATE_IDLE;
	
	reset_read_mode = FALSE;

#ifdef DEBUG
	comopen(DEBUG_COMPORT);
#endif

	DEBUGPUTSTRING( "\r\nApplication started\r\n" ) ;

	// The following function implements systemsetting menu options that aren't present in the OS or need to be overruled
	AddMenuOptionTable(menu_option_table, MAX_OPTIONS);

	//scannerpower(SINGLE | TRIGGER, -1);	// Initialize the scanner power to default and let the OS handle the rest

	// Enable remembering of (decoder) settings in flash and load old settings if present
	if (SystemSettingsMemorizing(ON) < 0) // Enable remembering of (decoder) settings in flash and load old settings if present
	{
		Application_Default(COM0); // If no old settings were found -> Reset OS and application to default
	}

	// Enable remembering of application settings in flash and load old settings if present and compatible
	if (ApplicationSettingsMemorizing(ON, (void *) &app, sizeof(app), VERSION_CHECK) < 0)
	{
		Application_Default(COM0); // If no valid application settings were found -> Reset to OS and application default
	}

	systemsetting("YC");	// Always makes sure menu-labels are enabled after a restart
	
	goodreadled(ORANGE, -1);
	OpenStorage();				// Init storage of barcodes (Create File if not exist)
	goodreadled(ORANGE, 0);
	
	// If TRUE, then the small clear key acts like a trigger key, so read modes settings also apply to this key
	ClearKeyAsTrigger((app.output_quantity_options & QUANTITY_DELETE_KEY_DISABLED) ? FALSE : TRUE);
	
	//
	// Disable all timers
	//
	for(int x = 0; x < MAX_TIMERS; ++x)				
		timer[x] = 0xFFFF;
	
	//
	// Install our own heartbeat timer that beats every 20ms
	//
	InstallHeartbeatHandler( Timers_Heartbeat );

	// Set shutdown callback to close the barcode database
	System_SetShutdownCallback(ShutdownCallback);
	
	//
	// We're controlling the read mode in the app unfortunately; set the last settings from the app vars
	//
	SetReadMode(app.read_mode, app.read_time);

	autopowerdown(ON, IDLE_POWER_DOWN_TIME);				// Init auto power down time
	autopowerdown(APD_SHUTDOWN_ON, IDLE_POWER_OFF_TIME);	// Init auto power off time

	if (app.buzzer_options & STARTUP_BUZZER)
    {
        sound(TSTANDARD, VHIGH, SLOW, SMEDIUM, SHIGH, SLOW, SMEDIUM, SHIGH, 0);	// Start-up beep
    }

    for (;;)
    {		
        //
        // Main state machine
        //
        switch (state)
        {
            case STATE_IDLE:		// Idle state (not connected)
                if ((ch = getchar()) > 0)
                    last_key_pressed = ch;// Used to determine whether a barcode was read using the clear key or trigger key
					
				if (downpressed() && app.func_key_del)
				{
					if (timer[FORMAT_TIMER] == 0xFFFF)
					{
						timer[FORMAT_TIMER] = CLR_KEY_FORMAT_TIME;
					}
					else if (timer[FORMAT_TIMER] == 0)
					{
						timer[FORMAT_TIMER] = 0xFFFF;
						goodreadled(GREEN, 100);
						remove(FNAME_SCANNED);
						remove(FNAME_SCANNED_EXPORT);
						remove(FNAME_SCANNED_IDX);		// Remove current link index file as well
						format();
						sound(TSTANDARD, VHIGH, SHIGH, SMEDIUM, SLOW, 0);
					}
				}
				else
				{
					timer[FORMAT_TIMER] = 0xFFFF;
				}

                //
                // Barcode reading and handling
                //
                if (readbarcode(&code) == OK)			// If barcode read
                {
                    ProcessBarcode(&code);				// Process the barcode
					reset_read_mode = TRUE;
				}
                
                if (UsbStatus() != USB_NOT_CONNECTED)	// If the USB is connected
                {
                    if (app.default_interface != COM11)	// If not MSD -> Generate the export file here. Else Generate it in the shutdown-callback
                    {
                        CloseStorage();					// Close storage
                        autopowerdown(SUSPEND, 0);
                        ExportDatabase();				// Export the database
                        autopowerdown(RESUME, 0);
						//goodreadled(GREEN_FLASH, -1);	// Show blinking green LED when connected to USB
					}
					
                    state = STATE_USB_CONNECTED;		// Switch the state to usb connected
                    //scannerpower(SUSPEND, 0);			// Remember the current read mode
                    scannerpower(OFF, 0);				// Disable scanning while waiting to transmit its data
                }
                break;

            case STATE_USB_CONNECTED:
                if (UsbStatus() == USB_NOT_CONNECTED)
                {
                    if (app.default_interface == COM11)	// USB-MSD: Powerdown on disconnecting the USB cable, because the file system might have changed completely the OS will restart the application!!!!!!!
                    {
                        state = STATE_WAIT_FOR_RELEASE;
						reset_read_mode = TRUE;
						break;
                    }

					goodreadled(ORANGE, -1);
                    OpenStorage();					// Re-open the database
					goodreadled(ORANGE, 0);

                    //scannerpower(RESUME, 0);		// Restore the read mode
					reset_read_mode = TRUE;
                    state = STATE_WAIT_FOR_RELEASE;	// Forces an update of the read mode, which may have been changed
                    break;
                }

                if (app.default_interface != COM11)
                {
                    Communicate();
                }
                break;

            case STATE_WAIT_FOR_RELEASE:
            default:
                if (!triggerpressed() && !downpressed())
                {
                    goodreadled(GREEN_FLASH, 0);// Stop the green flashing LED (if enabled)
                    state = STATE_IDLE;
                }
                break;
        }
		
		if ( reset_read_mode && !triggerpressed() )
        {
            reset_read_mode = FALSE;
            SetReadMode(app.read_mode, app.read_time);
        }
		
        idle();		// This reduces the power consumption
    }
}
